package sun.rmi.transport.tcp;

import java.rmi.server.LogStream;
import java.util.Enumeration;
import java.io.IOException;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import java.util.Hashtable;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.OutputStream;
import java.io.InputStream;
import sun.rmi.runtime.Log;

final class ConnectionMultiplexer
{
    static int logLevel;
    static final Log multiplexLog;
    private static final int OPEN = 225;
    private static final int CLOSE = 226;
    private static final int CLOSEACK = 227;
    private static final int REQUEST = 228;
    private static final int TRANSMIT = 229;
    private TCPChannel channel;
    private InputStream in;
    private OutputStream out;
    private boolean orig;
    private DataInputStream dataIn;
    private DataOutputStream dataOut;
    private Hashtable<Integer, MultiplexConnectionInfo> connectionTable;
    private int numConnections;
    private static final int maxConnections = 256;
    private int lastID;
    private boolean alive;
    
    private static String getLogLevel() {
        return (String)AccessController.doPrivileged(new GetPropertyAction("sun.rmi.transport.tcp.multiplex.logLevel"));
    }
    
    public ConnectionMultiplexer(TCPChannel channel, InputStream inputStream, OutputStream outputStream, boolean orig) {
        this.connectionTable = new Hashtable(7);
        this.numConnections = 0;
        this.lastID = 4097;
        this.alive = true;
        this.channel = channel;
        this.in = inputStream;
        this.out = outputStream;
        this.orig = orig;
        this.dataIn = new DataInputStream(inputStream);
        this.dataOut = new DataOutputStream(outputStream);
    }
    
    public void run() throws IOException {
        while (true) {
            try {
                int unsignedByte = this.dataIn.readUnsignedByte();
                switch (unsignedByte) {
                    case 225: {
                        int unsignedShort = this.dataIn.readUnsignedShort();
                        if (ConnectionMultiplexer.multiplexLog.isLoggable(Log.VERBOSE)) {
                            ConnectionMultiplexer.multiplexLog.log(Log.VERBOSE, "operation  OPEN " + unsignedShort);
                        }
                        if ((MultiplexConnectionInfo)this.connectionTable.get(Integer.valueOf(unsignedShort)) != null) {
                            throw new IOException("OPEN: Connection ID already exists");
                        }
                        MultiplexConnectionInfo value = new MultiplexConnectionInfo(unsignedShort);
                        value.in = new MultiplexInputStream(this, value, 2048);
                        value.out = new MultiplexOutputStream(this, value, 2048);
                        synchronized (this.connectionTable) {
                            this.connectionTable.put(Integer.valueOf(unsignedShort), value);
                            ++this.numConnections;
                        }
                        this.channel.acceptMultiplexConnection(new TCPConnection(this.channel, value.in, value.out));
                        continue;
                    }
                    case 226: {
                        int unsignedShort2 = this.dataIn.readUnsignedShort();
                        if (ConnectionMultiplexer.multiplexLog.isLoggable(Log.VERBOSE)) {
                            ConnectionMultiplexer.multiplexLog.log(Log.VERBOSE, "operation  CLOSE " + unsignedShort2);
                        }
                        MultiplexConnectionInfo multiplexConnectionInfo = (MultiplexConnectionInfo)this.connectionTable.get(Integer.valueOf(unsignedShort2));
                        if (multiplexConnectionInfo == null) {
                            throw new IOException("CLOSE: Invalid connection ID");
                        }
                        multiplexConnectionInfo.in.disconnect();
                        multiplexConnectionInfo.out.disconnect();
                        if (!multiplexConnectionInfo.closed) {
                            this.sendCloseAck(multiplexConnectionInfo);
                        }
                        synchronized (this.connectionTable) {
                            this.connectionTable.remove(Integer.valueOf(unsignedShort2));
                            --this.numConnections;
                        }
                        continue;
                    }
                    case 227: {
                        int unsignedShort3 = this.dataIn.readUnsignedShort();
                        if (ConnectionMultiplexer.multiplexLog.isLoggable(Log.VERBOSE)) {
                            ConnectionMultiplexer.multiplexLog.log(Log.VERBOSE, "operation  CLOSEACK " + unsignedShort3);
                        }
                        MultiplexConnectionInfo multiplexConnectionInfo2 = (MultiplexConnectionInfo)this.connectionTable.get(Integer.valueOf(unsignedShort3));
                        if (multiplexConnectionInfo2 == null) {
                            throw new IOException("CLOSEACK: Invalid connection ID");
                        }
                        if (!multiplexConnectionInfo2.closed) {
                            throw new IOException("CLOSEACK: Connection not closed");
                        }
                        multiplexConnectionInfo2.in.disconnect();
                        multiplexConnectionInfo2.out.disconnect();
                        synchronized (this.connectionTable) {
                            this.connectionTable.remove(Integer.valueOf(unsignedShort3));
                            --this.numConnections;
                        }
                        continue;
                    }
                    case 228: {
                        int unsignedShort4 = this.dataIn.readUnsignedShort();
                        MultiplexConnectionInfo multiplexConnectionInfo3 = (MultiplexConnectionInfo)this.connectionTable.get(Integer.valueOf(unsignedShort4));
                        if (multiplexConnectionInfo3 == null) {
                            throw new IOException("REQUEST: Invalid connection ID");
                        }
                        int int1 = this.dataIn.readInt();
                        if (ConnectionMultiplexer.multiplexLog.isLoggable(Log.VERBOSE)) {
                            ConnectionMultiplexer.multiplexLog.log(Log.VERBOSE, "operation  REQUEST " + unsignedShort4 + ": " + int1);
                        }
                        multiplexConnectionInfo3.out.request(int1);
                        continue;
                    }
                    case 229: {
                        int unsignedShort5 = this.dataIn.readUnsignedShort();
                        MultiplexConnectionInfo multiplexConnectionInfo4 = (MultiplexConnectionInfo)this.connectionTable.get(Integer.valueOf(unsignedShort5));
                        if (multiplexConnectionInfo4 == null) {
                            throw new IOException("SEND: Invalid connection ID");
                        }
                        int int2 = this.dataIn.readInt();
                        if (ConnectionMultiplexer.multiplexLog.isLoggable(Log.VERBOSE)) {
                            ConnectionMultiplexer.multiplexLog.log(Log.VERBOSE, "operation  TRANSMIT " + unsignedShort5 + ": " + int2);
                        }
                        multiplexConnectionInfo4.in.receive(int2, this.dataIn);
                        continue;
                    }
                    default: {
                        throw new IOException("Invalid operation: " + Integer.toHexString(unsignedByte));
                    }
                }
            }
            finally {
                this.shutDown();
            }
        }
    }
    
    public synchronized TCPConnection openConnection() throws IOException {
        int lastID;
        do {
            this.lastID = (++this.lastID & 0x7FFF);
            lastID = this.lastID;
            if (this.orig) {
                lastID |= 0x8000;
            }
            else {
                continue;
            }
        } while (this.connectionTable.get(Integer.valueOf(lastID)) != null);
        MultiplexConnectionInfo value = new MultiplexConnectionInfo(lastID);
        value.in = new MultiplexInputStream(this, value, 2048);
        value.out = new MultiplexOutputStream(this, value, 2048);
        synchronized (this.connectionTable) {
            if (!this.alive) {
                throw new IOException("Multiplexer connection dead");
            }
            if (this.numConnections >= 256) {
                throw new IOException("Cannot exceed 256 simultaneous multiplexed connections");
            }
            this.connectionTable.put(Integer.valueOf(lastID), value);
            ++this.numConnections;
        }
        synchronized (this.dataOut) {
            try {
                this.dataOut.writeByte(225);
                this.dataOut.writeShort(lastID);
                this.dataOut.flush();
            }
            catch (IOException ex) {
                ConnectionMultiplexer.multiplexLog.log(Log.BRIEF, "exception: ", ex);
                this.shutDown();
                throw ex;
            }
        }
        return new TCPConnection(this.channel, value.in, value.out);
    }
    
    public void shutDown() {
        synchronized (this.connectionTable) {
            if (!this.alive) {
                return;
            }
            this.alive = false;
            Enumeration<MultiplexConnectionInfo> elements = this.connectionTable.elements();
            while (elements.hasMoreElements()) {
                MultiplexConnectionInfo multiplexConnectionInfo = (MultiplexConnectionInfo)elements.nextElement();
                multiplexConnectionInfo.in.disconnect();
                multiplexConnectionInfo.out.disconnect();
            }
            this.connectionTable.clear();
            this.numConnections = 0;
        }
        try {
            this.in.close();
        }
        catch (IOException ex) {}
        try {
            this.out.close();
        }
        catch (IOException ex2) {}
    }
    
    void sendRequest(MultiplexConnectionInfo multiplexConnectionInfo, int v) throws IOException {
        synchronized (this.dataOut) {
            if (this.alive && !multiplexConnectionInfo.closed) {
                try {
                    this.dataOut.writeByte(228);
                    this.dataOut.writeShort(multiplexConnectionInfo.id);
                    this.dataOut.writeInt(v);
                    this.dataOut.flush();
                }
                catch (IOException ex) {
                    ConnectionMultiplexer.multiplexLog.log(Log.BRIEF, "exception: ", ex);
                    this.shutDown();
                    throw ex;
                }
            }
        }
    }
    
    void sendTransmit(MultiplexConnectionInfo multiplexConnectionInfo, byte[] b, int off, int n) throws IOException {
        synchronized (this.dataOut) {
            if (this.alive && !multiplexConnectionInfo.closed) {
                try {
                    this.dataOut.writeByte(229);
                    this.dataOut.writeShort(multiplexConnectionInfo.id);
                    this.dataOut.writeInt(n);
                    this.dataOut.write(b, off, n);
                    this.dataOut.flush();
                }
                catch (IOException ex) {
                    ConnectionMultiplexer.multiplexLog.log(Log.BRIEF, "exception: ", ex);
                    this.shutDown();
                    throw ex;
                }
            }
        }
    }
    
    void sendClose(MultiplexConnectionInfo multiplexConnectionInfo) throws IOException {
        multiplexConnectionInfo.out.disconnect();
        synchronized (this.dataOut) {
            if (this.alive && !multiplexConnectionInfo.closed) {
                try {
                    this.dataOut.writeByte(226);
                    this.dataOut.writeShort(multiplexConnectionInfo.id);
                    this.dataOut.flush();
                    multiplexConnectionInfo.closed = true;
                }
                catch (IOException ex) {
                    ConnectionMultiplexer.multiplexLog.log(Log.BRIEF, "exception: ", ex);
                    this.shutDown();
                    throw ex;
                }
            }
        }
    }
    
    void sendCloseAck(MultiplexConnectionInfo multiplexConnectionInfo) throws IOException {
        synchronized (this.dataOut) {
            if (this.alive && !multiplexConnectionInfo.closed) {
                try {
                    this.dataOut.writeByte(227);
                    this.dataOut.writeShort(multiplexConnectionInfo.id);
                    this.dataOut.flush();
                    multiplexConnectionInfo.closed = true;
                }
                catch (IOException ex) {
                    ConnectionMultiplexer.multiplexLog.log(Log.BRIEF, "exception: ", ex);
                    this.shutDown();
                    throw ex;
                }
            }
        }
    }
    
    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        this.shutDown();
    }
    
    static {
        ConnectionMultiplexer.logLevel = LogStream.parseLevel(getLogLevel());
        multiplexLog = Log.getLog("sun.rmi.transport.tcp.multiplex", "multiplex", ConnectionMultiplexer.logLevel);
    }
}
